/*
 * Copyright (C) 2004, 2005 Moritz Orbach <zufall@apfelboymchen.homeunix.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>

#ifdef _GNU_SOURCE
#include <unistd.h>
#endif

#include "config.h"

#include "list.h"
#include "status.h"

#ifdef _GNU_SOURCE
t_list *visited;
#endif

/* can I ask imlib2 for a list of supported filetypes? */
char *supported_filetypes[] = {
	".jpg",
	".jpeg",
	".png",
	".gif",
	".bmp",
	".lbm",
	".pnm",
	".tga",
	".targa",
	".tif",
	".tiff",
	".xpm",
	NULL
};

#ifdef DEBUG
void list_list(t_list *list)
{
	int i;

	for (i = 1; i <= list_count(list); i++)
		printf("%.3d: %s\n", i, (char *)list_get(list, i));

}
#endif

/*
 * returns true if the filetype is supported
 */
int supported(const char *file)
{
	char **s;
	char *ext;

	s = supported_filetypes;

	ext = rindex(file, '.');
	if (!ext)
		return 0;

	while (*s != NULL) {
		if (!strcasecmp(ext, *s))
			return 1;	/* supported */

		*s++;
	}
	/* nothing found */
	return 0;
}


/*
 * returns
 *    1 if path is a directory / a link to a directory
 *    0 if path is a file / a link to a file
 *   -1 on error
 */
int isdir(const char *path)
{
	int sr; /* stat return */
	struct stat estat;

	sr = stat(path, &estat);
	if (sr) {
		perror(path);
		return -1;
	}

	if (!S_ISDIR(estat.st_mode))
		return 0;
	else return 1;
}


#ifdef _GNU_SOURCE
/* symlink-loop-detection under GNU-systems */

int alreadythere(t_list *visited, char *path)
{
	int i;
	char *savedpath = NULL;

	for (i = 1; i <= list_count(visited); i++) {
		savedpath = list_get(visited, i);
		/*
		printf("vergleiche %s und %s\n", path, savedpath);
		*/
		if (!strcmp(path, savedpath))
			return 1;
	}
	return 0;
}

/*
 * returns
 *    0 - no loop
 *    1 - loop detected
 *   -1 - error
 */
int symloop(const char *path)
{
	extern t_list *visited;

	char *realtarget;
	int loop = 0;
	int dir = 0;

	dir = isdir(path);
	if (dir != 1)
		return dir;

	/*
	 * warning: implicit declaration of function `canonicalize_file_name'
	 * warning: assignment makes pointer from integer without a cast
	 * Why?!
	 */
	realtarget = canonicalize_file_name(path);
	if(realtarget == NULL) {
		perror(path);
		return -1;
	}

	if (alreadythere(visited, realtarget))
		loop = 1;
	
	free(realtarget);
	return loop;
}
#endif

/*
 * adds a file / the files in the directory recursivly to the list
 */
void addpath(const char *path)
{
	extern t_list *imglist;
	struct dirent *dp;
	DIR *dirp;

	switch(isdir(path)) {
		case -1: break; /* error */
		case  0: /* path is a file or link to a file - add it to the list */
			 if (supported(path))
				 list_append(imglist, path, strlen(path) + 1);
			 break;
		case  1: /* path is a directory or a link to a directory */
#ifdef _GNU_SOURCE
			 if (chdir(path) != 0) {
				 perror("chdir");
				 break;
			 }

			 /* remember it */
			 list_append(visited, path, strlen(path) + 1);
#endif

			 dirp = opendir(path);
			 if (dirp == NULL) {
				 /* error or file? */
				 perror(path);
				 return;
			 }

			 /* reading directory entries */
			 while (running() && (dp = readdir (dirp)) != NULL) {

				 /* cat directory + last component of path (2 ISO C90-errors) */
				 char fullpath[strlen(path) + strlen(dp->d_name) + 2]; /* +1 \0, +1 '/'*/
				 strcpy(fullpath, path);
				 /* append slash if there is none */
				 if (fullpath[strlen(fullpath) - 1] != '/')
					 strncat(fullpath, "/", 1);
				 strcat(fullpath, dp->d_name);


				 /* DT_* isn't available with -ansi */
				 if (dp->d_type == DT_LNK) {
				 /* link to a file OR a directory */
#ifdef _GNU_SOURCE
					 /* This needs further inspection */
					 char *cpath;
					 cpath = canonicalize_file_name(fullpath);
					 if (cpath == NULL) {
						 printf("%s: dead symlink: %s\n", PACKAGE, fullpath);
						 break;
					 }
					 switch (symloop(cpath)) {
						 case 0: addpath(cpath);
							 free(cpath);
							 break;
						 case 1: /* hmm, this dir looks familiar */
							 /*
							 fprintf(stderr, "symlink-loop detected: %s -> %s\n", fullpath, cpath);
							  */
							 fprintf(stderr, "symlink-loop detected\n");
							 free(cpath);
							 break;
						 default:
							 /* else error */
							 free(cpath);
							 break;
					 }
#else
					 /* no risk, no fun */
					 addpath(fullpath);
#endif
				 } else if (dp->d_type == DT_DIR) {
					 /* directory */
					 if (!strcmp(dp->d_name, ".")
							 || !strcmp(dp->d_name, "..")
							 || !strcmp(dp->d_name, ".xvpics")) /* go away .xvpics! we now have ~/.thumbnails/! */
						 continue;

					 addpath(fullpath);

				 } else if (dp->d_type == DT_REG) {
					 /* file */
					 if (supported(fullpath))
						 list_append(imglist, fullpath, strlen(fullpath) + 1);

				 }
			 }
			 closedir(dirp);
			 break;
	}
}


int buildlist(t_list *imglist, int argc, char *argv[])
{
#ifdef _GNU_SOURCE
	int i;
	char *temp;
	char *iwd;

	visited = list_init();
	iwd = (char *)get_current_dir_name();
	for (i = 0; i < argc; i++) {
		temp = canonicalize_file_name(argv[i]);
		if (temp) {
			addpath(temp);
			free(temp);
		} else
			perror(argv[i]);
		chdir(iwd);
	}
	free(iwd);
	list_destroy(visited);
#else
	int i;
	for (i = 0; i < argc; i++)
		addpath(argv[i]);
#endif

	return list_count(imglist);
}
